"""
MIT License

Copyright (c) 2023-2024 AstreaTSS

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import typing
from datetime import datetime
from enum import IntEnum
from uuid import UUID

import msgspec
import typing_extensions as typing_ext

from elytra.core import CamelBaseModel, ParsableCamelModel, add_decoder

__all__ = (
    "ClubUserPresence",
    "ClubDeeplinkMetadata",
    "ClubDeeplinks",
    "ClubPresence",
    "ClubType",
    "ProfileMetadata",
    "Profile",
    "TitleDeeplinkMetadata",
    "TitleDeeplinks",
    "Club",
    "ClubResponse",
)


def _camel_to_const_snake(s: str) -> str:
    return "".join([f"_{c}" if c.isupper() else c.upper() for c in s]).lstrip("_")


class ClubUserPresence(IntEnum):
    UNKNOWN = -1
    NOT_IN_CLUB = 0
    IN_CLUB = 1
    CHAT = 2
    FEED = 3
    ROSTER = 4
    PLAY = 5
    IN_GAME = 6

    @classmethod
    def from_xbox_api(cls, value: str) -> typing_ext.Self:
        try:
            return cls[_camel_to_const_snake(value)]
        except KeyError:
            # it's not like i forgot a value, it's just that some are
            # literally not documented
            return cls.UNKNOWN


class ClubDeeplinkMetadata(CamelBaseModel):
    page_name: str
    uri: str


class ClubDeeplinks(CamelBaseModel):
    xbox: list[ClubDeeplinkMetadata]
    pc: list[ClubDeeplinkMetadata]


class ClubPresence(CamelBaseModel):
    xuid: str
    last_seen_timestamp: datetime
    _last_seen_state: str = msgspec.field(name="lastSeenState")

    @property
    def last_seen_state(self) -> ClubUserPresence:
        return ClubUserPresence.from_xbox_api(self._last_seen_state)


class ClubType(CamelBaseModel):
    type: str
    genre: str
    localized_title_family_name: str
    title_family_id: UUID


class ProfileMetadata(CamelBaseModel):
    can_viewer_change_setting: bool
    value: typing.Optional[typing.Any] = None
    allowed_values: typing.Optional[typing.Any] = None


class Profile(CamelBaseModel):
    description: ProfileMetadata
    rules: ProfileMetadata
    name: ProfileMetadata
    short_name: ProfileMetadata
    is_searchable: ProfileMetadata
    is_recommendable: ProfileMetadata
    request_to_join_enabled: ProfileMetadata
    open_join_enabled: ProfileMetadata
    leave_enabled: ProfileMetadata
    transfer_ownership_enabled: ProfileMetadata
    mature_content_enabled: ProfileMetadata
    watch_club_titles_only: ProfileMetadata
    display_image_url: ProfileMetadata
    background_image_url: ProfileMetadata
    preferred_locale: ProfileMetadata
    tags: ProfileMetadata
    associated_titles: ProfileMetadata
    primary_color: ProfileMetadata
    secondary_color: ProfileMetadata
    tertiary_color: ProfileMetadata


class TitleDeeplinkMetadata(CamelBaseModel):
    title_id: str
    uri: str = msgspec.field(name="Uri")


class TitleDeeplinks(CamelBaseModel):
    xbox: list[TitleDeeplinkMetadata]
    pc: list[TitleDeeplinkMetadata]
    android: list[TitleDeeplinkMetadata]
    ios: list[TitleDeeplinkMetadata] = msgspec.field(
        name="iOS"
    )  # this makes more sense at least


class Club(CamelBaseModel):
    id: str
    club_type: ClubType
    creation_date_utc: datetime
    glyph_image_url: str
    banner_image_url: str
    followers_count: int
    members_count: int
    moderators_count: int
    recommended_count: int
    requested_to_join_count: int
    club_presence_count: int
    club_presence_today_count: int
    club_presence_in_game_count: int
    club_presence: list[ClubPresence]
    state: str
    report_count: int
    reported_items_count: int
    max_members_per_club: int
    max_members_in_game: int
    founder_xuid: str
    title_deeplinks: TitleDeeplinks
    profile: Profile
    club_deeplinks: ClubDeeplinks
    owner_xuid: typing.Optional[str] = None  # ???
    suspended_until_utc: typing.Optional[typing.Any] = None
    roster: typing.Optional[typing.Any] = None
    target_roles: typing.Optional[typing.Any] = None
    recommendation: typing.Optional[typing.Any] = None
    settings: typing.Optional[typing.Any] = None
    short_name: typing.Optional[typing.Any] = None


@add_decoder
class ClubResponse(ParsableCamelModel):
    clubs: list[Club]
    search_facet_results: typing.Optional[typing.Any] = None
    recommendation_counts: typing.Optional[typing.Any] = None
    club_deeplinks: typing.Optional[typing.Any] = None



"""
MIT License

Copyright (c) 2023-2024 AstreaTSS

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import typing
from datetime import datetime

from elytra.core import CamelBaseModel, ParsableCamelModel, PascalBaseModel, add_decoder

__all__ = (
    "PeopleSummaryResponse",
    "Suggestion",
    "Recommendation",
    "MultiplayerSummary",
    "RecentPlayer",
    "Follower",
    "PreferredColor",
    "PresenceDetail",
    "TitlePresence",
    "Detail",
    "SocialManager",
    "Avatar",
    "LinkedAccount",
    "Person",
    "RecommendationSummary",
    "FriendFinderState",
    "PeopleHubResponse",
)


class PeopleSummaryResponse(CamelBaseModel):
    target_following_count: int
    target_follower_count: int
    is_caller_following_target: bool
    is_target_following_caller: bool
    has_caller_marked_target_as_favorite: bool
    has_caller_marked_target_as_identity_shared: bool
    legacy_friend_status: str
    available_people_slots: typing.Optional[int] = None
    recent_change_count: typing.Optional[int] = None
    watermark: typing.Optional[str] = None


# microsoft is a great and consistent company
class Suggestion(PascalBaseModel):
    priority: int
    type: typing.Optional[str] = None
    reasons: typing.Optional[str] = None
    title_id: typing.Optional[str] = None


class Recommendation(PascalBaseModel):
    type: str
    reasons: list[str]


class MultiplayerSummary(PascalBaseModel):
    in_multiplayer_session: int
    in_party: int


class RecentPlayer(CamelBaseModel):
    titles: list[str]
    text: typing.Optional[str] = None


class Follower(CamelBaseModel):
    text: typing.Optional[str] = None
    followed_date_time: typing.Optional[datetime] = None


class PreferredColor(CamelBaseModel):
    primary_color: typing.Optional[str] = None
    secondary_color: typing.Optional[str] = None
    tertiary_color: typing.Optional[str] = None


class PresenceDetail(PascalBaseModel):
    is_broadcasting: bool
    device: str
    presence_text: str
    state: str
    title_id: str
    is_primary: bool
    is_game: bool
    title_type: typing.Optional[str] = None
    rich_presence_text: typing.Optional[str] = None


class TitlePresence(PascalBaseModel):
    is_currently_playing: bool
    presence_text: typing.Optional[str] = None
    title_name: typing.Optional[str] = None
    title_id: typing.Optional[str] = None


class Detail(CamelBaseModel):
    account_tier: str
    is_verified: bool
    watermarks: list[str]
    blocked: bool
    mute: bool
    follower_count: int
    following_count: int
    has_game_pass: bool
    bio: typing.Optional[str] = None
    location: typing.Optional[str] = None
    tenure: typing.Optional[str] = None


class SocialManager(CamelBaseModel):
    title_ids: list[str]
    pages: list[str]


class Avatar(CamelBaseModel):
    update_time_offset: typing.Optional[datetime] = None
    spritesheet_metadata: typing.Optional[typing.Any] = None


class LinkedAccount(CamelBaseModel):
    network_name: str
    show_on_profile: bool
    is_family_friendly: bool
    display_name: typing.Optional[str] = None
    deeplink: typing.Optional[str] = None


class Person(CamelBaseModel):
    xuid: str
    is_favorite: bool
    is_following_caller: bool
    is_followed_by_caller: bool
    is_identity_shared: bool
    real_name: str
    display_pic_raw: str
    show_user_as_avatar: str
    gamertag: str
    gamer_score: str
    modern_gamertag: str
    modern_gamertag_suffix: str
    unique_modern_gamertag: str
    xbox_one_rep: str
    presence_state: str
    presence_text: str
    color_theme: str
    preferred_flag: str
    is_broadcasting: bool
    preferred_platforms: list[str]
    is_quarantined: bool
    is_xbox360_gamerpic: bool
    presence_devices: typing.Optional[typing.Any] = None
    is_cloaked: typing.Optional[bool] = None
    added_date_time_utc: typing.Optional[datetime] = None
    display_name: typing.Optional[str] = None
    suggestion: typing.Optional[Suggestion] = None
    recommendation: typing.Optional[Recommendation] = None
    search: typing.Optional[typing.Any] = None
    title_history: typing.Optional[typing.Any] = None
    multiplayer_summary: typing.Optional[MultiplayerSummary] = None
    recent_player: typing.Optional[RecentPlayer] = None
    follower: typing.Optional[Follower] = None
    preferred_color: typing.Optional[PreferredColor] = None
    presence_details: typing.Optional[list[PresenceDetail]] = None
    title_presence: typing.Optional[TitlePresence] = None
    title_summaries: typing.Optional[typing.Any] = None
    presence_title_ids: typing.Optional[list[str]] = None
    detail: typing.Optional[Detail] = None
    community_manager_titles: typing.Optional[typing.Any] = None
    social_manager: typing.Optional[SocialManager] = None
    broadcast: typing.Optional[list[typing.Any]] = None
    tournament_summary: typing.Optional[typing.Any] = None
    avatar: typing.Optional[Avatar] = None
    linked_accounts: typing.Optional[list[LinkedAccount]] = None
    last_seen_date_time_utc: typing.Optional[datetime] = None


class RecommendationSummary(CamelBaseModel):
    friend_of_friend: int
    facebook_friend: int
    phone_contact: int
    follower: int
    VIP: int
    steam_friend: int
    promote_suggestions: bool


class FriendFinderState(CamelBaseModel):
    facebook_opt_in_status: str
    facebook_token_status: str
    phone_opt_in_status: str
    phone_token_status: str
    steam_opt_in_status: str
    steam_token_status: str
    discord_opt_in_status: str
    discord_token_status: str
    instagram_opt_in_status: str
    instagram_token_status: str
    mixer_opt_in_status: str
    mixer_token_status: str
    reddit_opt_in_status: str
    reddit_token_status: str
    twitch_opt_in_status: str
    twitch_token_status: str
    twitter_opt_in_status: str
    twitter_token_status: str
    you_tube_opt_in_status: str
    you_tube_token_status: str


@add_decoder
class PeopleHubResponse(ParsableCamelModel):
    people: list[Person]
    recommendation_summary: typing.Optional[RecommendationSummary] = None
    friend_finder_state: typing.Optional[FriendFinderState] = None
    account_link_details: typing.Optional[list[LinkedAccount]] = None



"""
MIT License

Copyright (c) 2023-2024 AstreaTSS

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from elytra.const import BEDROCK_REALMS_API_URL, MC_VERSION
from elytra.core import BaseMicrosoftAPI

from .models import *

__all__ = (
    "Permission",
    "State",
    "WorldType",
    "FullRealm",
    "MultiRealmResponse",
    "Player",
    "PartialRealm",
    "ActivityListResponse",
    "RealmCountResponse",
    "PendingInvite",
    "PendingInviteResponse",
    "BedrockRealmsAPI",
)


class BedrockRealmsAPI(BaseMicrosoftAPI):
    RELYING_PATH: str = BEDROCK_REALMS_API_URL
    BASE_URL: str = BEDROCK_REALMS_API_URL

    @property
    def base_headers(self) -> dict[str, str]:
        return {
            "Authorization": self.auth_mgr.xsts_token.authorization_header_value,
            "Client-Version": MC_VERSION,
            "User-Agent": "MCPE/UWP",
        }

    async def join_realm_from_code(self, code: str) -> FullRealm:
        return await FullRealm.from_response(
            await self.post(f"invites/v1/link/accept/{code}")
        )

    async def fetch_realms(self) -> MultiRealmResponse:
        return await MultiRealmResponse.from_response(await self.get("worlds"))

    async def invite_player(
        self, realm_id: str | int, player_xuid: str | int
    ) -> FullRealm:
        return await FullRealm.from_response(
            await self.put(
                f"invites/{realm_id}/invite/update",
                json={"invites": {str(player_xuid): "ADD"}},
            )
        )

    async def fetch_pending_invite_count(self) -> int:
        resp = await self.get("invites/count/pending")
        return int(await resp.aread())

    async def fetch_pending_invites(self) -> PendingInviteResponse:
        return await PendingInviteResponse.from_response(
            await self.get("invites/pending")
        )

    async def accept_invite(self, invitation_id: str) -> None:
        await self.put(f"invites/accept/{invitation_id}")

    async def reject_invite(self, invitation_id: str) -> None:
        await self.put(f"invites/reject/{invitation_id}")

    async def fetch_activities(self) -> ActivityListResponse:
        return await ActivityListResponse.from_response(
            await self.get("activities/live/players")
        )

    async def leave_realm(self, realm_id: int | str) -> None:
        await self.delete(f"invites/{realm_id}")

    async def fetch_realm_count(self) -> RealmCountResponse:
        return await RealmCountResponse.from_response(
            await self.get("clubs/membercount")
        )



"""
MIT License

Copyright (c) 2023-2024 AstreaTSS

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import contextlib
import functools
import inspect
import secrets
import traceback
import typing
from enum import IntEnum

import anyio
from anyio.streams.stapled import StapledObjectStream
from anyio.streams.tls import TLSStream
from websockets.client import ClientProtocol
from websockets.frames import Frame, Opcode
from websockets.uri import parse_uri

__all__ = ("RTAType", "RTA")

try:
    import orjson

    def _loads_wrapper(json_input: str | bytes) -> typing.Any:
        return orjson.loads(json_input)

except ImportError:
    import msgspec

    decoder = msgspec.json.Decoder()

    def _loads_wrapper(json_input: str | bytes) -> typing.Any:
        return decoder.decode(json_input)


def _generate_id() -> bytes:
    return secrets.token_bytes()


class RTAType(IntEnum):
    SUBSCRIBE = 1
    UNSUBSCRIBE = 2
    EVENT = 3
    RSYNC = 4


PING_INTERVAL = 20.0
PING_TIMEOUT = 20.0


class RTA:
    def __init__(self, headers: dict | None = None) -> None:
        self._uri = parse_uri("wss://rta.xboxlive.com/connect")
        self._headers = headers or {}
        self._protocol = ClientProtocol(self._uri)
        self._stream: TLSStream | None = None

        self._tg = anyio.create_task_group()
        self._exit_stack = contextlib.AsyncExitStack()

        self._last_sequence_number = 0
        self._endpoint_maps: dict[
            int, typing.Callable[[list[typing.Any]], typing.Awaitable[typing.Any]]
        ] = {}
        self._subscribe_listeners: dict[
            int, typing.Callable[[list[typing.Any]], typing.Awaitable[typing.Any]]
        ] = {}
        self._stapled_stream_set: set[StapledObjectStream] = set()

        self._ping_tasks: dict[bytes, anyio.Event] = {}

    @classmethod
    async def establish(cls, headers: dict | None = None) -> typing.Self:
        self = cls(headers=headers)
        await self.connect()
        return self

    async def connect(self) -> None:
        self._stream = await anyio.connect_tcp(
            self._uri.host, self._uri.port, tls=True, tls_standard_compatible=False
        )

        request = self._protocol.connect()
        request.headers.update(self._headers)
        self._protocol.send_request(request)

        for data in self._protocol.data_to_send():
            await self._stream.send(data)

        try:
            data = await self._stream.receive()
        except anyio.EndOfStream:
            self._protocol.receive_eof()
            for data in self._protocol.data_to_send():
                await self._stream.send(data)

            raise ConnectionError("Connection closed.") from None

        self._protocol.receive_data(data)

        if self._protocol.handshake_exc is not None:
            raise self._protocol.handshake_exc

        await self._exit_stack.enter_async_context(self._tg)
        self._tg.start_soon(self._receive)
        self._tg.start_soon(self._send_ping)

    async def _receive(self) -> None:
        if not self._stream:
            raise ConnectionError("Stream hasn't been started yet.")

        try:
            while True:
                try:
                    data = await self._stream.receive()
                except anyio.EndOfStream:
                    self._protocol.receive_eof()
                    for data in self._protocol.data_to_send():
                        await self._stream.send(data)

                    await self.close()
                    raise ConnectionError("Received EOF.") from None
                except anyio.ClosedResourceError:
                    return

                self._protocol.receive_data(data)

                if self._protocol.handshake_exc is not None:
                    await self.close()
                    raise self._protocol.handshake_exc

                events = self._protocol.events_received()

                for event in events:
                    if isinstance(event, Frame):
                        if event.opcode == Opcode.CLOSE:
                            await self.close()
                            return

                        if event.opcode == Opcode.PING:
                            self._protocol.send_pong(event.data)
                            for data in self._protocol.data_to_send():
                                await self._stream.send(data)

                        elif (
                            event.opcode == Opcode.PONG
                            and bytes(event.data) in self._ping_tasks
                        ):
                            self._ping_tasks[bytes(event.data)].set()

                        elif event.opcode in {Opcode.BINARY, Opcode.TEXT}:
                            await self._handle_data(event.data)

        except anyio.get_cancelled_exc_class():
            return

        except Exception as e:
            traceback.print_exception(e)
            raise e

    async def _send_ping(self) -> None:
        if not self._stream:
            raise ConnectionError("Stream hasn't been started yet.")

        try:
            while True:
                await anyio.sleep(PING_INTERVAL)

                ping_id = _generate_id()
                self._ping_tasks[ping_id] = anyio.Event()

                self._protocol.send_ping(ping_id)
                for data in self._protocol.data_to_send():
                    await self._stream.send(data)

                with anyio.fail_after(PING_TIMEOUT):
                    await self._ping_tasks[ping_id].wait()
                    self._ping_tasks.pop(ping_id)

        except TimeoutError:
            print("Ping timeout.")  # noqa: T201
            await self.close()

        except anyio.get_cancelled_exc_class():
            return

        except Exception as e:
            traceback.print_exception(e)
            raise e

    async def _send_str(self, data: str) -> None:
        if not self._stream:
            raise ConnectionError("Connection closed.")

        self._protocol.send_text(data.encode())
        for data_to_send in self._protocol.data_to_send():
            await self._stream.send(data_to_send)

    async def _handle_data(self, data: bytes) -> None:
        if not self._stream:
            raise ConnectionError("Connection closed.")

        parsed_data: list[typing.Any] = _loads_wrapper(data)

        if parsed_data[0] == RTAType.SUBSCRIBE:
            await self._subscribe_listeners[parsed_data[1]](parsed_data)
        elif parsed_data[0] == RTAType.EVENT:
            self._tg.start_soon(self._endpoint_maps[parsed_data[1]], parsed_data)

    async def subscribe(
        self,
        url: str,
        dispatch_handler: typing.Callable[
            [list[typing.Any]], typing.Awaitable[typing.Any]
        ],
    ) -> None:
        if not inspect.iscoroutinefunction(dispatch_handler):
            raise ValueError("dispatch_handler must be a coroutine function.")

        url = url.removesuffix("/")
        self._last_sequence_number += 1
        to_send = f'[{RTAType.SUBSCRIBE},{self._last_sequence_number},"{url}"]'

        stapled_stream = StapledObjectStream(*anyio.create_memory_object_stream())

        self._subscribe_listeners[self._last_sequence_number] = functools.partial(
            self._subscribe_handle, dispatch_handler, stapled_stream
        )
        await self._send_str(to_send)

        self._stapled_stream_set.add(stapled_stream)

        try:
            data = await stapled_stream.receive()
        except (anyio.ClosedResourceError, anyio.EndOfStream):
            return

        self._stapled_stream_set.discard(stapled_stream)

        if isinstance(data, Exception):
            raise data

    async def unsubscribe(self, subscription_id: int) -> None:
        self._last_sequence_number += 1
        to_send = (
            f"[{RTAType.UNSUBSCRIBE},{self._last_sequence_number}, {subscription_id}]"
        )
        await self._send_str(to_send)
        self._endpoint_maps.pop(subscription_id, None)

    async def _subscribe_handle(
        self,
        new_dispatch_handle: typing.Callable[
            [list[typing.Any]], typing.Awaitable[typing.Any]
        ],
        stapled_stream: StapledObjectStream,
        data: list[typing.Any],
    ) -> None:
        if len(data) != 5:
            if len(data) == 4:
                await stapled_stream.send(ValueError(f"Invalid RTA: {data[3]}"))
            else:
                await stapled_stream.send(ValueError(f"Invalid RTA: {data}"))
            return

        self._endpoint_maps[data[3]] = new_dispatch_handle
        self._subscribe_listeners.pop(data[1], None)

        await stapled_stream.send(None)

    async def close(self) -> None:
        if self._stream is not None:
            await self._stream.aclose()
            self._stream = None

        for stream in self._stapled_stream_set:
            await stream.aclose()

        self._tg.cancel_scope.cancel()
        await self._exit_stack.aclose()



"""
MIT License

Copyright (c) 2023-2024 AstreaTSS

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime
import functools
import typing

import httpx
import msgspec
import typing_extensions as typing_ext

from elytra import retry_transport
from elytra.protocols import HandlerProtocol

__all__ = (
    "BaseModel",
    "CamelBaseModel",
    "PascalBaseModel",
    "ParsableModel",
    "ParsableCamelModel",
    "ParsablePascalModel",
    "add_decoder",
    "OAuth2TokenResponse",
    "AuthenticationManager",
    "MicrosoftAPIException",
    "BaseMicrosoftAPI",
)


def utc_now() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc)


class ParsableBase:
    if typing.TYPE_CHECKING:
        _decoder: msgspec.json.Decoder[typing_ext.Self]
        _encoded_fields: frozenset[str]

    @classmethod
    def from_data(cls, obj: dict) -> typing_ext.Self:
        if not hasattr(cls, "_encoded_fields"):
            cls._encoded_fields = frozenset(
                f.encode_name for f in msgspec.structs.fields(cls)
            )

        filted_dict = {k: v for k, v in obj.items() if k in cls._encoded_fields}
        return cls(**filted_dict)

    @classmethod
    def from_bytes(cls, obj: bytes) -> typing_ext.Self:
        return cls._decoder.decode(obj)  # type: ignore

    @classmethod
    async def from_response(cls, resp: httpx.Response) -> typing_ext.Self:
        return cls.from_bytes(await resp.aread())


class ParsableModel(msgspec.Struct, ParsableBase, kw_only=True):
    pass


class ParsableCamelModel(msgspec.Struct, ParsableBase, rename="camel", kw_only=True):
    pass


class ParsablePascalModel(msgspec.Struct, ParsableBase, rename="pascal", kw_only=True):
    pass


PM = typing.TypeVar("PM", bound=type[ParsableBase])


def add_decoder(cls: PM) -> PM:
    cls._decoder = msgspec.json.Decoder(cls)
    return cls


class BaseModel(msgspec.Struct, kw_only=True):
    pass


class CamelBaseModel(msgspec.Struct, rename="camel", kw_only=True):
    pass


class PascalBaseModel(msgspec.Struct, rename="pascal", kw_only=True):
    pass


@add_decoder
class OAuth2TokenResponse(ParsableModel):
    token_type: str
    expires_in: int
    scope: str
    access_token: str
    issued: datetime.datetime = msgspec.field(default_factory=utc_now)
    user_id: typing.Optional[str] = None
    refresh_token: typing.Optional[str] = None

    def is_valid(self) -> bool:
        return (self.issued + datetime.timedelta(seconds=self.expires_in)) > utc_now()

    @classmethod
    def from_file(cls, path: str) -> typing_ext.Self:
        with open(path, "rb") as f:
            return cls.from_bytes(f.read())


class XTokenResponse(ParsablePascalModel):
    issue_instant: datetime.datetime
    not_after: datetime.datetime
    token: str

    def is_valid(self) -> bool:
        return self.not_after > utc_now()


class DisplayClaims(BaseModel):
    xui: list[dict[str, str]]


@add_decoder
class XAUResponse(XTokenResponse):
    display_claims: DisplayClaims


@add_decoder
class XSTSResponse(XTokenResponse):
    display_claims: DisplayClaims

    @property
    def xuid(self) -> str:
        return self.display_claims.xui[0]["xid"]

    @property
    def userhash(self) -> str:
        return self.display_claims.xui[0]["uhs"]

    @property
    def gamertag(self) -> str:
        return self.display_claims.xui[0]["gtg"]

    @property
    def age_group(self) -> str:
        return self.display_claims.xui[0]["agg"]

    @property
    def privileges(self) -> str:
        return self.display_claims.xui[0]["prv"]

    @property
    def user_privileges(self) -> str:
        return self.display_claims.xui[0]["usr"]

    @property
    def authorization_header_value(self) -> str:
        return f"XBL3.0 x={self.userhash};{self.token}"


class AuthenticationManager:
    __slots__ = (
        "session",
        "client_id",
        "client_secret",
        "relying_party",
        "oauth",
        "user_token",
        "xsts_token",
    )

    session: httpx.AsyncClient
    client_id: str
    client_secret: str
    relying_party: str

    oauth: OAuth2TokenResponse
    user_token: XAUResponse
    xsts_token: XSTSResponse

    def __init__(
        self,
        session: httpx.AsyncClient,
        client_id: str,
        client_secret: str,
        relying_party: str,
    ) -> None:
        self.session = session
        self.client_id = client_id
        self.client_secret = client_secret
        self.relying_party = relying_party

        self.oauth = None  # type: ignore
        self.user_token = None  # type: ignore
        self.xsts_token = None  # type: ignore

    @classmethod
    async def from_ouath(
        cls,
        session: httpx.AsyncClient,
        client_id: str,
        client_secret: str,
        relying_party: str,
        oauth: OAuth2TokenResponse,
    ) -> typing_ext.Self:
        self = cls(session, client_id, client_secret, relying_party)
        self.oauth = oauth
        await self.refresh_tokens()
        return self

    @classmethod
    async def from_data(
        cls,
        session: httpx.AsyncClient,
        client_id: str,
        client_secret: str,
        relying_party: str,
        oauth_data: dict,
    ) -> typing_ext.Self:
        self = cls(session, client_id, client_secret, relying_party)
        self.oauth = OAuth2TokenResponse.from_data(oauth_data)
        await self.refresh_tokens()
        return self

    @classmethod
    async def from_file(
        cls,
        session: httpx.AsyncClient,
        client_id: str,
        client_secret: str,
        relying_party: str,
        oauth_path: str,
    ) -> typing_ext.Self:
        self = cls(session, client_id, client_secret, relying_party)
        self.oauth = OAuth2TokenResponse.from_file(oauth_path)
        await self.refresh_tokens()
        return self

    async def _oauth2_token_request(self, data: dict) -> OAuth2TokenResponse:
        """Execute token requests."""
        data["client_id"] = self.client_id
        if self.client_secret:
            data["client_secret"] = self.client_secret
        resp = await self.session.post(
            "https://login.live.com/oauth20_token.srf", data=data
        )
        resp.raise_for_status()
        return await OAuth2TokenResponse.from_response(resp)

    async def request_oauth_token(
        self, authorization_code: str, redirect_uri: str
    ) -> OAuth2TokenResponse:
        """Request OAuth2 token."""
        return await self._oauth2_token_request(
            {
                "grant_type": "authorization_code",
                "code": authorization_code,
                "scope": "Xboxlive.signin Xboxlive.offline_access",
                "redirect_uri": redirect_uri,
            }
        )

    async def refresh_oauth_token(self) -> OAuth2TokenResponse:
        """Refresh OAuth2 token."""
        if not self.oauth.refresh_token:
            raise ValueError("No refresh token present.")

        return await self._oauth2_token_request(
            {
                "grant_type": "refresh_token",
                "scope": "Xboxlive.signin Xboxlive.offline_access",
                "refresh_token": self.oauth.refresh_token,
            },
        )

    async def request_user_token(
        self,
        relying_party: str = "http://auth.xboxlive.com",
        use_compact_ticket: bool = False,
    ) -> XAUResponse:
        """Authenticate via access token and receive user token."""
        url = "https://user.auth.xboxlive.com/user/authenticate"
        headers = {"x-xbl-contract-version": "1"}
        data = {
            "RelyingParty": relying_party,
            "TokenType": "JWT",
            "Properties": {
                "AuthMethod": "RPS",
                "SiteName": "user.auth.xboxlive.com",
                "RpsTicket": (
                    self.oauth.access_token
                    if use_compact_ticket
                    else f"d={self.oauth.access_token}"
                ),
            },
        }

        resp = await self.session.post(url, json=data, headers=headers)
        resp.raise_for_status()
        return await XAUResponse.from_response(resp)

    async def request_xsts_token(
        self,
    ) -> XSTSResponse:
        """Authorize via user token and receive final X token."""
        url = "https://xsts.auth.xboxlive.com/xsts/authorize"
        headers = {"x-xbl-contract-version": "1"}
        data = {
            "RelyingParty": self.relying_party,
            "TokenType": "JWT",
            "Properties": {
                "UserTokens": [self.user_token.token],
                "SandboxId": "RETAIL",
            },
        }

        resp = await self.session.post(url, json=data, headers=headers)
        resp.raise_for_status()
        return await XSTSResponse.from_response(resp)

    async def refresh_tokens(self, force_refresh: bool = False) -> None:
        """Refresh all tokens."""
        if force_refresh:
            self.oauth = await self.refresh_oauth_token()
            self.user_token = await self.request_user_token()
            self.xsts_token = await self.request_xsts_token()
        else:
            if not (self.oauth and self.oauth.is_valid()):
                self.oauth = await self.refresh_oauth_token()
            if not (self.user_token and self.user_token.is_valid()):
                self.user_token = await self.request_user_token()
            if not (self.xsts_token and self.xsts_token.is_valid()):
                self.xsts_token = await self.request_xsts_token()

    async def close(self) -> None:
        await self.session.aclose()


class MicrosoftAPIException(Exception):
    def __init__(self, resp: httpx.Response, error: Exception) -> None:
        self.resp = resp
        self.error = error

        super().__init__(
            "An error occured when trying to access this resource: code"
            f" {resp.status_code}.\nError: {error}"
        )


try:
    import orjson

    def _dumps_wrapper(obj: typing.Any) -> bytes:
        return orjson.dumps(obj)

except ImportError:
    encoder = msgspec.json.Encoder()

    def _dumps_wrapper(obj: typing.Any) -> bytes:
        return encoder.encode(obj)


async def should_retry(
    auth_mgr: AuthenticationManager, response: httpx.Response
) -> bool:
    if not response.is_error:
        return False

    if response.status_code == 401:
        await auth_mgr.refresh_tokens(force_refresh=True)
        return True

    return True


class BaseMicrosoftAPI(HandlerProtocol):
    RELYING_PATH: str = "http://xboxlive.com"
    BASE_URL: str = ""

    session: httpx.AsyncClient
    auth_mgr: AuthenticationManager

    def __init__(
        self, session: httpx.AsyncClient, auth_mgr: AuthenticationManager
    ) -> None:
        self.session = session
        self.auth_mgr = auth_mgr

    @classmethod
    async def from_oauth(
        cls, client_id: str, client_secret: str, oauth: OAuth2TokenResponse
    ) -> typing_ext.Self:
        transport = retry_transport.RetryTransport(
            wrapped_transport=httpx.AsyncHTTPTransport(http2=True, retries=2),
            jitter_ratio=0.3,
        )
        session = httpx.AsyncClient(transport=transport)
        auth_mgr = await AuthenticationManager.from_ouath(
            session, client_id, client_secret, cls.RELYING_PATH, oauth
        )

        session._transport._should_retry_async = functools.partial(should_retry, auth_mgr)  # type: ignore
        return cls(session, auth_mgr)

    @classmethod
    async def from_data(
        cls, client_id: str, client_secret: str, oauth_data: dict
    ) -> typing_ext.Self:
        transport = retry_transport.RetryTransport(
            wrapped_transport=httpx.AsyncHTTPTransport(http2=True, retries=2),
            jitter_ratio=0.3,
        )
        session = httpx.AsyncClient(transport=transport)
        auth_mgr = await AuthenticationManager.from_data(
            session, client_id, client_secret, cls.RELYING_PATH, oauth_data
        )

        session._transport._should_retry_async = functools.partial(should_retry, auth_mgr)  # type: ignore
        return cls(session, auth_mgr)

    @classmethod
    async def from_file(
        cls, client_id: str, client_secret: str, oauth_path: str = "oauth.json"
    ) -> typing_ext.Self:
        transport = retry_transport.RetryTransport(
            wrapped_transport=httpx.AsyncHTTPTransport(http2=True, retries=2),
            jitter_ratio=0.3,
        )
        session = httpx.AsyncClient(transport=transport)
        auth_mgr = await AuthenticationManager.from_file(
            session, client_id, client_secret, cls.RELYING_PATH, oauth_path
        )

        session._transport._should_retry_async = functools.partial(should_retry, auth_mgr)  # type: ignore
        return cls(session, auth_mgr)

    @property
    def base_headers(self) -> dict[str, str]:
        return {"Authorization": self.auth_mgr.xsts_token.authorization_header_value}

    async def close(self) -> None:
        await self.session.aclose()

    async def request(
        self,
        method: str,
        url: str,
        json: typing.Any = None,
        data: typing.Optional[dict] = None,
        params: typing.Optional[dict] = None,
        headers: typing.Optional[dict] = None,
        *,
        force_refresh: bool = False,
        dont_handle_ratelimit: bool = False,
        **kwargs: typing.Any,
    ) -> httpx.Response:
        if not headers:
            headers = {}
        if not params:
            params = {}

        # refresh token as needed
        await self.auth_mgr.refresh_tokens(force_refresh=force_refresh)

        if json:
            if data:
                raise ValueError("Cannot use both json and data.")

            kwargs["content"] = _dumps_wrapper(json)
            headers["Content-Type"] = "application/json"

        req_kwargs = {
            "method": method,
            "url": f"{self.BASE_URL}{url}",
            "headers": headers | self.base_headers,
            "data": data,
            "params": params,
        } | kwargs

        if dont_handle_ratelimit:
            req_kwargs["extensions"] = {"dont_handle_ratelimit": True}

        resp = await self.session.request(**req_kwargs)

        try:
            resp.raise_for_status()
            return resp
        except Exception as e:
            raise MicrosoftAPIException(resp, e) from e

    async def get(
        self,
        url: str,
        json: typing.Any = None,
        data: typing.Optional[dict] = None,
        params: typing.Optional[dict] = None,
        headers: typing.Optional[dict] = None,
        **kwargs: typing.Any,
    ) -> httpx.Response:
        return await self.request(
            "GET",
            url,
            json=json,
            data=data,
            params=params,
            headers=headers,
            **kwargs,
        )

    async def post(
        self,
        url: str,
        json: typing.Any = None,
        data: typing.Optional[dict] = None,
        params: typing.Optional[dict] = None,
        headers: typing.Optional[dict] = None,
        **kwargs: typing.Any,
    ) -> httpx.Response:
        return await self.request(
            "POST",
            url,
            json=json,
            data=data,
            params=params,
            headers=headers,
            **kwargs,
        )

    async def put(
        self,
        url: str,
        json: typing.Any = None,
        data: typing.Optional[dict] = None,
        params: typing.Optional[dict] = None,
        headers: typing.Optional[dict] = None,
        **kwargs: typing.Any,
    ) -> httpx.Response:
        return await self.request(
            "PUT",
            url,
            json=json,
            data=data,
            params=params,
            headers=headers,
            **kwargs,
        )

    async def delete(
        self,
        url: str,
        json: typing.Any = None,
        data: typing.Optional[dict] = None,
        params: typing.Optional[dict] = None,
        headers: typing.Optional[dict] = None,
        **kwargs: typing.Any,
    ) -> httpx.Response:
        return await self.request(
            "DELETE",
            url,
            json=json,
            data=data,
            params=params,
            headers=headers,
            **kwargs,
        )



"""
MIT License

Copyright (c) 2023-2024 AstreaTSS

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime
import typing
from enum import Enum
from types import NoneType

import msgspec

from elytra.core import CamelBaseModel, ParsableCamelModel, add_decoder

__all__ = (
    "Permission",
    "State",
    "WorldType",
    "FullRealm",
    "MultiRealmResponse",
    "Player",
    "PartialRealm",
    "ActivityListResponse",
    "PendingInvite",
    "PendingInviteResponse",
    "RealmCountResponse",
)


class Permission(Enum):
    VISITOR = "VISITOR"
    MEMBER = "MEMBER"
    OPERATOR = "OPERATOR"


class State(Enum):
    CLOSED = "CLOSED"
    OPEN = "OPEN"


class WorldType(Enum):
    NORMAL = "NORMAL"


class Player(CamelBaseModel):
    uuid: str
    name: NoneType
    operator: bool
    accepted: bool
    online: bool
    permission: Permission


@add_decoder
class FullRealm(ParsableCamelModel):
    id: int
    remote_subscription_id: str
    owner: typing.Optional[str]
    name: str
    default_permission: Permission
    state: State
    days_left: int
    expired: bool
    expired_trial: bool
    grace_period: bool
    world_type: WorldType
    max_players: int
    minigame_name: NoneType
    minigame_id: NoneType
    minigame_image: NoneType
    active_slot: int
    member: bool
    subscription_refresh_status: NoneType
    owner_uuid: str = msgspec.field(name="ownerUUID")
    slots: typing.Optional[typing.Any] = None
    players: typing.Optional[list[Player]] = None
    club_id: typing.Optional[int] = None
    motd: typing.Optional[str] = None


@add_decoder
class MultiRealmResponse(ParsableCamelModel):
    servers: list[FullRealm]


class PartialRealm(CamelBaseModel):
    id: int
    players: list[Player]
    full: bool


@add_decoder
class ActivityListResponse(ParsableCamelModel):
    servers: list[PartialRealm]


class PendingInvite(CamelBaseModel):
    invitation_id: str  # not the realm id, funny enough
    world_name: str
    world_description: str
    world_owner_name: str
    world_owner_uuid: str
    date_timestamp: int = msgspec.field(name="date")

    @property
    def date(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(self.date_timestamp, tz=datetime.UTC)


@add_decoder
class PendingInviteResponse(ParsableCamelModel):
    invites: list[PendingInvite]


@add_decoder
class RealmCountResponse(ParsableCamelModel):
    realm_count: int = msgspec.field(name="memberCount")
    realm_limit: int = msgspec.field(name="memberLimit")
    over_limit: bool



"""
MIT License

Copyright (c) 2023-2024 AstreaTSS

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse
import contextlib
import os
from email.utils import formatdate
from urllib.parse import parse_qsl, urlencode

import anyio
import httpx
import msgspec
from anyio.abc import SocketStream

from elytra import AuthenticationManager

code: str = ""


def generate_authorization_url(client_id: str, redirect_uri: str) -> str:
    """Generate Windows Live Authorization URL."""
    query_params = {
        "client_id": client_id,
        "response_type": "code",
        "approval_prompt": "auto",
        "scope": "Xboxlive.signin Xboxlive.offline_access",
        "redirect_uri": redirect_uri,
    }

    return f"https://login.live.com/oauth20_authorize.srf?{urlencode(query_params)}"


async def respond_to_request(client: SocketStream, response_status: str) -> None:
    resp = (
        f"HTTP/1.1 {response_status}\r\nDate: {formatdate(usegmt=True)}\r\nServer:"
        " elytra-ms\r\n\r\n"
    )
    await client.send(resp.encode())


async def respond_with_text(
    client: SocketStream, response_status: str, text: str
) -> None:
    resp = (
        f"HTTP/1.1 {response_status}\r\nContent-Type: text/plain;"
        f" charset=utf-8\r\nContent-Length: {len(text)}\r\nDate:"
        f" {formatdate(usegmt=True)}\r\nServer: elytra-ms\r\n\r\n{text}"
    )
    await client.send(resp.encode())


async def handle_microsoft(client: SocketStream) -> None:
    global code

    async with client:
        raw_data = await client.receive()

        try:
            data = raw_data.decode()
        except UnicodeDecodeError:
            return await respond_to_request(client, "400 Bad Request")

        if not data.startswith("GET /auth/callback?code="):
            return await respond_to_request(client, "400 Bad Request")

        first_couple_of_entries = data.split(maxsplit=2)

        url_ish = first_couple_of_entries[1]
        query_string = url_ish.split("?", 1)[1]
        query_dict = dict(parse_qsl(query_string))

        if query_dict.get("error"):
            await respond_with_text(
                client,
                "400 Bad Request",
                f"Error: {query_dict.get('error_description')}",
            )
        elif gotten_code := query_dict.get("code"):
            code = gotten_code
            await respond_with_text(
                client,
                "200 OK",
                "Success! You may now close this tab and go back to the terminal.",
            )
        else:
            await respond_with_text(
                client, "400 Bad Request", "Error: No code was provided."
            )

        await client.aclose()
        raise anyio.EndOfStream


async def async_main() -> None:
    global code

    parser = argparse.ArgumentParser(description="Authenticate into Microsoft's APIs.")
    parser.add_argument(
        "--file",
        "-f",
        default="oauth.json",
        help="Filepath to put output of program. Default: 'oauth.json'.",
    )
    parser.add_argument(
        "--client-id",
        "-cid",
        default=os.environ.get("CLIENT_ID"),
        help="OAuth2 client ID.",
    )
    parser.add_argument(
        "--client-secret",
        default=os.environ.get("CLIENT_SECRET"),
        help="OAuth2 client secret.",
    )
    args = parser.parse_args()

    if not (args.client_id and args.client_secret):
        raise ValueError(
            "You must provide a client ID and secret either via the CLI or environment"
            " variables."
        )

    redirect_uri = "http://localhost:8080/auth/callback"

    print(  # noqa: T201
        f"Please visit {generate_authorization_url(args.client_id, redirect_uri)} to"
        " authenticate."
    )

    listener = await anyio.create_tcp_listener(local_host="localhost", local_port=8080)

    with contextlib.suppress(anyio.EndOfStream):
        await listener.serve(handle_microsoft)

    if not code:
        print("Authentication failed.")  # noqa: T201
        return

    auth_mgr = AuthenticationManager(
        httpx.AsyncClient(http2=True),
        args.client_id,
        args.client_secret,
        "http://xboxlive.com",
    )
    oauth = await auth_mgr.request_oauth_token(code, redirect_uri)
    with open(args.file, mode="wb") as f:
        f.write(msgspec.json.encode(oauth))

    await auth_mgr.close()
    print("Authentication successful!")  # noqa: T201


def main() -> None:
    anyio.run(async_main)


if __name__ == "__main__":
    main()



# SKIP LICENSE INSERTION

"""
elytra-ms
~~~~~~~~~~~~~~~~~~~

A Python Library for various Microsoft APIs, including the Xbox and Bedrock Realms APIs.

(c) 2023-2024 AstreaTSS under MIT, see LICENSE for more details.
"""

from .bedrock_realms import *
from .const import *
from .core import *
from .protocols import *
from .xbox import *

__version__ = "0.3.2"



"""
MIT License

Copyright (c) 2023-2024 AstreaTSS

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse
import os

import anyio
import httpx
import msgspec


async def async_main() -> None:
    parser = argparse.ArgumentParser(description="Authenticate into Microsoft's APIs.")
    parser.add_argument(
        "--file",
        "-f",
        default="oauth.json",
        help="Filepath to put output of program. Default: 'oauth.json'.",
    )
    parser.add_argument(
        "--client-id",
        "-cid",
        default=os.environ.get("CLIENT_ID"),
        help="OAuth2 client ID.",
    )
    args = parser.parse_args()

    if not args.client_id:
        raise ValueError(
            "You must provide a client ID either via the CLI or environment variables."
        )

    session = httpx.AsyncClient()

    r = await session.post(
        "https://login.microsoftonline.com/consumers/oauth2/v2.0/devicecode",
        data={
            "client_id": args.client_id,
            "scope": "Xboxlive.signin Xboxlive.offline_access",
        },
    )
    data = await r.json()
    if r.is_error:
        print(data)  # noqa: T201
        return

    success_response: dict | None = None

    print(data["message"])  # noqa: T201

    try:
        with anyio.fail_after(data["expires_in"]):
            while True:
                await anyio.sleep(data["interval"])
                r = await session.post(
                    "https://login.microsoftonline.com/consumers/oauth2/v2.0/token",
                    data={
                        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                        "client_id": args.client_id,
                        "device_code": data["device_code"],
                    },
                )
                resp_json = await r.json()
                if error := resp_json.get("error"):
                    if error in {
                        "authorization_declined",
                        "expired_token",
                        "bad_verification_code",
                    }:
                        break
                else:
                    success_response = resp_json
                    break
    except TimeoutError:
        print("Authentication timed out.")  # noqa: T201
        return

    if not success_response:
        print("Authentication failed.")  # noqa: T201
        return

    with open(args.file, "wb") as f:
        f.write(msgspec.json.encode(success_response))

    print("Authentication successful!")  # noqa: T201


def main() -> None:
    anyio.run(async_main)


if __name__ == "__main__":
    main()
